@@ -1,6 +1,6 @@
 import os,sys
-# os.chdir(os.path.dirname(os.path.abspath(__file__)))
-# sys.path.insert(1, os.path.join(sys.path[0], '..'))
+#os.chdir(os.path.dirname(os.path.abspath(__file__)))
+#sys.path.insert(1, os.path.join(sys.path[0], '..'))
 
 import requests
 import pandas as pd
@@ -9,7 +9,7 @@ import random
 import datetime,time
 import logging
 import re
-import urllib.parse
+import urllib.parse 
 
 mode ='vpn'
 
@@ -26,11 +26,20 @@ if(mode=='vpn'):
         except ValueError:  # includes simplejson.decoder.JSONDecodeError:
             payload2 = "https://www.nseindia.com"
             output2 = os.popen('curl -c cookies.txt "'+payload2+'"'+curl_headers+'').read()
-
+    
             output = os.popen(payload_var).read()
             output=json.loads(output)
         return output
-
+if(mode=='local'):
+    def nsefetch(payload):
+        try:
+            output = requests.get(payload,headers=headers).json()
+            print(output)
+        except ValueError:
+            s =requests.Session()
+            output = s.get("http://nseindia.com",headers=headers)
+            output = s.get(payload,headers=headers).json()
+        return output
 
 
 headers = {
@@ -158,7 +167,7 @@ def oi_chain_builder(symbol,expiry="latest",oi_mode="full"):
 
 
 def nse_quote(symbol,section=""):
-    #https://forum.unofficed.com/t/nsetools-get-quote-is-not-fetching-delivery-data-and-delivery-can-you-include-this-as-part-of-feature-request/1115/4
+    #https://forum.unofficed.com/t/nsetools-get-quote-is-not-fetching-delivery-data-and-delivery-can-you-include-this-as-part-of-feature-request/1115/4    
     symbol = nsesymbolpurify(symbol)
 
     if(section==""):
@@ -167,14 +176,18 @@ def nse_quote(symbol,section=""):
         else:
             payload = nsefetch('https://www.nseindia.com/api/quote-equity?symbol='+symbol)
         return payload
-
+    
     if(section!=""):
-        payload = nsefetch('https://www.nseindia.com/api/quote-equity?symbol='+symbol+'&section='+section)
+        payload = nsefetch('https://www.nseindia.com/api/quote-equity?symbol='+symbol+'&section='+section)            
         return payload
 
 
-def nse_expirydetails(payload,i=0):
-    currentExpiry = payload['records']['expiryDates'][i]
+def nse_expirydetails(payload,i=0):  #Can make problem. Use nse_expirydetails_by_symbol()
+
+    expiry_dates = payload['records']['expiryDates']
+    expiry_dates = [datetime.datetime.strptime(date, "%d-%b-%Y").date() for date in expiry_dates]
+    expiry_dates = [date.strftime("%d-%b-%Y") for date in expiry_dates if date >= datetime.datetime.now().date()]
+    currentExpiry=expiry_dates[i]    
     currentExpiry = datetime.datetime.strptime(currentExpiry,'%d-%b-%Y').date()  # converting json datetime to alice datetime
     date_today = run_time.strftime('%Y-%m-%d')  # required to remove hh:mm:ss
     date_today = datetime.datetime.strptime(date_today,'%Y-%m-%d').date()
@@ -193,29 +206,30 @@ def pcr(payload,inp='0'):
                 pass
     return pe_oi / ce_oi
 
+#forum.unofficed.com/t/unable-to-find-nse-quote-meta-api/702/4
+#Refer https://forum.unofficed.com/t/changed-the-nse-quote-ltp-function/1276
 def nse_quote_ltp(symbol,expiryDate="latest",optionType="-",strikePrice=0):
   payload = nse_quote(symbol)
-  #https://stackoverflow.com/questions/7961363/removing-duplicates-in-lists
-  #https://stackoverflow.com/questions/19199984/sort-a-list-in-python
-
-  #BankNIFTY and NIFTY has weekly options. Using this Jugaad which has primary base of assumption that Reliance will not step out of FNO.
-  #forum.unofficed.com/t/unable-to-find-nse-quote-meta-api/702/4
-  if((symbol in indices) and (optionType=="Fut")):
-    dates = expiry_list("RELIANCE","list")
-    if(expiryDate=="latest"): expiryDate=dates[0]
-    if(expiryDate=="next"): expiryDate=dates[1]
-
-  if(expiryDate=="latest") or (expiryDate=="next"):
-    dates=list(set((payload["expiryDates"])))
-    dates.sort(key = lambda date: datetime.datetime.strptime(date, '%d-%b-%Y'))
-    if(expiryDate=="latest"): expiryDate=dates[0]
-    if(expiryDate=="next"): expiryDate=dates[1]
 
   meta = "Options"
   if(optionType=="Fut"): meta = "Futures"
   if(optionType=="PE"):optionType="Put"
   if(optionType=="CE"):optionType="Call"
 
+  if(expiryDate=="latest") or (expiryDate=="next"):
+
+    if(meta=="Futures"):
+      selected_key = next((key for key in payload["expiryDatesByInstrument"] if "futures" in key.lower()), None)
+    if(meta=="Options"):
+      selected_key = next((key for key in payload["expiryDatesByInstrument"] if "options" in key.lower()), None)
+
+    expiry_dates=payload["expiryDatesByInstrument"][selected_key]
+    expiry_dates = [datetime.datetime.strptime(date, "%d-%b-%Y").date() for date in expiry_dates]
+    expiry_dates = [date.strftime("%d-%b-%Y") for date in expiry_dates if date >= datetime.datetime.now().date()]
+    if(expiryDate=="latest"): expiryDate=expiry_dates[0]
+    if(expiryDate=="next"): expiryDate=expiry_dates[1]
+  
+
   if(optionType!="-"):
       for i in payload['stocks']:
         if meta in i['metadata']['instrumentType']:
@@ -290,7 +304,10 @@ def nse_quote_meta(symbol,expiryDate="latest",optionType="-",strikePrice=0):
   return metadata
 
 def nse_optionchain_ltp(payload,strikePrice,optionType,inp=0,intent=""):
-    expiryDate=payload['records']['expiryDates'][inp]
+    expiry_dates = payload['records']['expiryDates']
+    expiry_dates = [datetime.datetime.strptime(date, "%d-%b-%Y").date() for date in expiry_dates]
+    expiry_dates = [date.strftime("%d-%b-%Y") for date in expiry_dates if date >= datetime.datetime.now().date()]
+    expiryDate=expiry_dates[inp]
     for x in range(len(payload['records']['data'])):
       if((payload['records']['data'][x]['strikePrice']==strikePrice) & (payload['records']['data'][x]['expiryDate']==expiryDate)):
           if(intent==""): return payload['records']['data'][x][optionType]['lastPrice']
@@ -593,13 +610,16 @@ def derivative_history_virgin(symbol,start_date,end_date,instrumentType,expiry_d
     instrumentType = instrumentType.lower()
 
     if(instrumentType=="options"):
-        if("NIFTY" in symbol): instrumentType="FUTSTK"
         instrumentType="OPTSTK"
-    if(instrumentType=="futures"):
         if("NIFTY" in symbol): instrumentType="OPTIDX"
-        instrumentType="FUTIDX"
+        
+    if(instrumentType=="futures"):
+        instrumentType="FUTSTK"
+        if("NIFTY" in symbol): instrumentType="FUTIDX"
+        
 
-    if(((instrumentType=="OPTIDX")or (instrumentType=="OPTSTK")) and (expiry_date!="")):
+    #if(((instrumentType=="OPTIDX")or (instrumentType=="OPTSTK")) and (expiry_date!="")):
+    if(strikePrice!=""):
         strikePrice = "%.2f" % strikePrice
         strikePrice = str(strikePrice)
 
@@ -667,11 +687,41 @@ def derivative_history(symbol,start_date,end_date,instrumentType,expiry_date,str
     return payload
 
 
-def expiry_history(symbol,start_date="",end_date=""):
+def expiry_history(symbol,start_date="",end_date="",type="options"):
     if(end_date==""):end_date=end_date
     nsefetch_url = "https://www.nseindia.com/api/historical/fo/derivatives/meta?&from="+start_date+"&to="+end_date+"&symbol="+symbol+""
     payload = nsefetch(nsefetch_url)
-    return payload['data'][2]
+
+    print(payload)
+
+    for key, value in payload['expiryDatesByInstrument'].items():
+      if type.lower() == "options" and "OPT" in key:
+          payload_data = payload['expiryDatesByInstrument'][key]
+          break
+      elif type.lower() == "futures" and "FUT" in key:
+          payload_data =  payload['expiryDatesByInstrument'][key]
+          break
+    
+    # Convert start_date and end_date to datetime objects
+    start_date = datetime.datetime.strptime(start_date, "%d-%m-%Y")
+    end_date = datetime.datetime.strptime(end_date, "%d-%m-%Y")
+
+    # Initialize an empty list to store filtered dates
+    filtered_date_payload = []
+
+    # Initialize a flag to check if the first date after end_date has been added
+    added_after_end_date = False    
+
+    # Iterate through date_payload and filter dates within the range
+    for date_str in payload_data:
+        date_obj = datetime.datetime.strptime(date_str, "%d-%b-%Y")
+        if start_date <= date_obj <= end_date:
+            filtered_date_payload.append(date_str)
+        elif date_obj > end_date and not added_after_end_date:
+            filtered_date_payload.append(date_str)
+            added_after_end_date = True
+    
+    return filtered_date_payload
 
 # # Nifty Indicies Site
 
@@ -693,22 +743,22 @@ niftyindices_headers = {
 }
 
 def index_history(symbol,start_date,end_date):
-    data = "{'name':'"+symbol+"','startDate':'"+start_date+"','endDate':'"+end_date+"'}"
-    payload = requests.post('https://niftyindices.com/Backpage.aspx/getHistoricaldatatabletoString', headers=niftyindices_headers,  data=data).json()
+    data = {'cinfo': "{'name':'" + symbol + "','startDate':'" + start_date + "','endDate':'" + end_date + "','indexName':'" + symbol + "'}"}
+    payload = requests.post('https://niftyindices.com/Backpage.aspx/getHistoricaldatatabletoString', headers=niftyindices_headers,  json=data).json()
     payload = json.loads(payload["d"])
     payload=pd.DataFrame.from_records(payload)
     return payload
 
 def index_pe_pb_div(symbol,start_date,end_date):
-    data = "{'name':'"+symbol+"','startDate':'"+start_date+"','endDate':'"+end_date+"'}"
-    payload = requests.post('https://niftyindices.com/Backpage.aspx/getpepbHistoricaldataDBtoString', headers=niftyindices_headers,  data=data).json()
+    data = {'cinfo': "{'name':'" + symbol + "','startDate':'" + start_date + "','endDate':'" + end_date + "','indexName':'" + symbol + "'}"}
+    payload = requests.post('https://niftyindices.com/Backpage.aspx/getpepbHistoricaldataDBtoString', headers=niftyindices_headers,  json=data).json()
     payload = json.loads(payload["d"])
     payload=pd.DataFrame.from_records(payload)
     return payload
 
 def index_total_returns(symbol,start_date,end_date):
-    data = "{'name':'"+symbol+"','startDate':'"+start_date+"','endDate':'"+end_date+"'}"
-    payload = requests.post('https://niftyindices.com/Backpage.aspx/getTotalReturnIndexString', headers=niftyindices_headers,  data=data).json()
+    data = {'cinfo': "{'name':'" + symbol + "','startDate':'" + start_date + "','endDate':'" + end_date + "','indexName':'" + symbol + "'}"}
+    payload = requests.post('https://niftyindices.com/Backpage.aspx/getTotalReturnIndexString', headers=niftyindices_headers,  json=data).json()
     payload = json.loads(payload["d"])
     payload=pd.DataFrame.from_records(payload)
     return payload
@@ -807,3 +857,79 @@ def nse_eq_symbols():
     #https://forum.unofficed.com/t/feature-request-stocklist-api/1073/11
     eq_list_pd = pd.read_csv('https://archives.nseindia.com/content/equities/EQUITY_L.csv')
     return eq_list_pd['SYMBOL'].tolist()
+
+def nse_price_band_hitters(bandtype="both",view="AllSec"):
+  payload = nsefetch("https://www.nseindia.com/api/live-analysis-price-band-hitter")
+  
+  #bandtype can be upper, lower, both
+  #view can be AllSec,SecGtr20,SecLwr20
+  return pd.DataFrame(payload[bandtype][view]["data"])
+
+def nse_largedeals(mode="bulk_deals"):
+  payload = nsefetch('https://www.nseindia.com/api/snapshot-capital-market-largedeal')
+  if(mode=="bulk_deals"):
+    return pd.DataFrame(payload["BULK_DEALS_DATA"])
+  if(mode=="short_deals"):
+    return pd.DataFrame(payload["SHORT_DEALS_DATA"])
+  if(mode=="block_deals"):
+    return pd.DataFrame(payload["BLOCK_DEALS_DATA"])
+
+def nse_largedeals_historical(from_date, to_date, mode="bulk_deals"):
+    if mode == "bulk_deals":
+        mode = "bulk-deals"
+    elif mode == "short_deals":
+        mode = "short-selling"
+    elif mode == "block_deals":
+        mode = "block-deals"
+    
+    url='https://www.nseindia.com/api/historical/' + mode + '?from=' + from_date + '&to=' + to_date
+    logging.info("Fetching " + str(url))
+    payload = nsefetch(url)
+    return pd.DataFrame(payload["data"])
+
+#https://forum.unofficed.com/t/feature-request-nse-fno-participant-wise-oi/1179/7
+#print(get_fao_participant_oi("04-06-2021"))
+def get_fao_participant_oi(date):
+    date = date.replace("-","")
+    payload=pd.read_csv("https://archives.nseindia.com/content/nsccl/fao_participant_oi_"+date+".csv")
+    return payload
+
+#https://forum.unofficed.com/t/how-to-check-if-the-market-is-open-today-or-not/1268/1
+def is_market_open(segment = "FO"): #COM,CD,CB,CMOT,COM,FO,IRD,MF,NDM,NTRP,SLBS
+    
+    holiday_json = nse_holidays()[segment]
+
+    # Get today's date in the format 'dd-Mon-yyyy'
+    today_date = datetime.date.today().strftime('%d-%b-%Y')
+
+    # Check if today's date is in the holiday_json
+    for holiday in holiday_json:
+        if holiday['tradingDate'] != today_date:
+            print("FNO Market is open today. Have a Nice Trade!")
+            return True
+        if holiday['tradingDate'] == today_date:
+            print(f"Market is closed today because of {holiday['description']}")
+            return False
+
+def nse_expirydetails_by_symbol(symbol,meta ="Futures",i=0):
+    payload = nse_quote(symbol)
+
+    if(meta=="Futures"):
+      selected_key = next((key for key in payload["expiryDatesByInstrument"] if "futures" in key.lower()), None)
+    if(meta=="Options"):
+      selected_key = next((key for key in payload["expiryDatesByInstrument"] if "options" in key.lower()), None)
+
+    expiry_dates=payload["expiryDatesByInstrument"][selected_key]
+    expiry_dates = [datetime.datetime.strptime(date, "%d-%b-%Y").date() for date in expiry_dates]
+    expiry_dates = [date.strftime("%d-%b-%Y") for date in expiry_dates if date >= datetime.datetime.now().date()]
+    
+    currentExpiry=expiry_dates[i]
+    currentExpiry = datetime.datetime.strptime(currentExpiry,'%d-%b-%Y').date()    
+    dte = (currentExpiry - datetime.datetime.now().date()).days
+    return currentExpiry,dte
+
+def security_wise_archive(from_date, to_date, symbol, series="ALL"):   
+    base_url = "https://www.nseindia.com/api/historical/securityArchives"
+    url = f"{base_url}?from={from_date}&to={to_date}&symbol={symbol.upper()}&dataType=priceVolumeDeliverable&series={series.upper()}"
+    payload = nsefetch(url)
+    return pd.DataFrame(payload['data'])
\ No newline at end of file
